from typing import List, Optional, Dict, Tuple, Any
import subprocess
import warnings


def call_subprocess(cmd_tokens: List[str],
                    cwd: str,
                    check: bool = True,
                    shell: bool = False,
                    env: Optional[Dict[str, str]] = None,
                    verbose: bool = False) -> Tuple[int, Any, Any]:
    """
    Executes a subprocess call.

    Parameters
    ----------
    cmd_tokens: list
        List of command tokens, e.g., ['ls', '-la']
    cwd: str
        Current working directory
    check: bool, optional
        Raise exception if command fails
    shell: bool, optional
        Run as shell command (not recommended)
    env: dict, optional
        Environment variables to pass to the subprocess
    verbose: bool, optional
        Flag for verbose output

    Returns
    -------
    Tuple[int, Any, Any]
        Decoded stdout of called process after completing

    Raises
    ------
    subprocess.CalledProcessError
        If an error occurs in the subprocess
    """

    try:
        r = subprocess.run(cmd_tokens, cwd=cwd, stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                           check=check, shell=shell, env=env)
        return r.returncode, r.stdout, r.stderr
    except subprocess.CalledProcessError as err:
        if verbose:
            warnings.warn(f"An error occurred in a subprocess call:\ncmd: {' '.join(cmd_tokens)}\n"
                          f"code: {err.returncode}\n"
                          f"output: {err.stdout} \nerror: {err.stderr}")
        return err.returncode, err.stdout, err.stderr



import os
from getpass import getpass
from typing import List, Any, Tuple
from .utils import call_subprocess


def _git(cmd_tokens: List[str],
         cwd: str) -> Tuple[int, Any, Any]:
    """
    Execute a git subprocess call.

    Parameters
    ----------
    cmd_tokens: list
        List of command tokens, e.g., ['ls', '-la']
    cwd: str
        Current working directory

    Returns
    -------
    Tuple
        Decoded stdout of called process after completing

    Raises
    ------
    subprocess.CalledProcessError
        If there is an error in the subprocess
    """

    env_vars = os.environ
    return call_subprocess(['git'] + cmd_tokens, cwd, check=True, shell=False, env=dict(env_vars), verbose=True)


class Git:

    def __init__(self,
                 repo_remote_url: str,
                 repo_local_directory: str,
                 cwd: str):
        """
        Load and configure bridge to a git repository.

        Parameters
        ----------
        repo_remote_url: str
            Remote URL of the git repository
        repo_local_directory: str
            Local directory to maintain the git repository
        cwd: str
            Current working directory
        """

        self.repo_local_directory = repo_local_directory
        self.repo_remote_url = repo_remote_url
        self.cwd = cwd

        print(f'Git bridge: {self.repo_local_directory} -- {self.repo_remote_url}')
        print(f'Git bridge cwd: {self.cwd}')

        if not os.path.isdir(self.repo_local_directory):
            res_code, stdout, err = self.clone()
            if res_code != 0:
                raise Exception(f'An error occurred while cloning the repository: {err.decode()}')
        else:
            print(f'Found existing repo in sync folder: {self.repo_local_directory}')

        print(f'Bridge initialized, bridge directory: {self.repo_local_directory}')

    def clone(self) -> Tuple[int, Any, Any]:
        """
        Clone the Overleaf git repository.

        Returns
        -------
        Tuple[int, Any, Any]
            the output from the git command
        """

        print('Cloning Overleaf git repo to sync.')
        res_code, stdout, err = _git(['clone', self.repo_remote_url, self.repo_local_directory], self.cwd)

        if res_code != 0:
            raise Exception(err)

        return res_code, stdout, err

    def add(self,
            file: str) -> Tuple[int, Any, Any]:
        """
        Add changes to the git repository.

        Parameters
        ----------
        file: str
            File to add to the staging

        Returns
        -------
        Tuple[int, Any, Any]
            the output from the git command
        """

        res_code, stdout, err = _git(['add', file], self.repo_local_directory)

        if res_code != 0:
            raise Exception(err)

        return res_code, stdout, err

    def commit(self) -> Tuple[int, Any, Any]:
        """
        Commit changes to the git repository.

        Returns
        -------
        Tuple[int, Any, Any]
            the output from the git command
        """

        res_code, stdout, err = _git(['commit', '-m', f'Update from Mizuna'], self.repo_local_directory)

        if res_code != 0:
            raise Exception(err)

        return res_code, stdout, err

    def pull(self) -> Tuple[int, Any, Any]:
        """
        Pull changes from the git repository.

        Returns
        -------
        Tuple[int, Any, Any]
            the output from the git command
        """

        res_code, stdout, err = _git(['pull'], self.repo_local_directory)

        if res_code != 0:
            raise Exception(err)

        return res_code, stdout, err

    def push(self) -> Tuple[int, Any, Any]:
        """
        Push changes to the git repository.

        Returns
        -------
        Tuple[int, Any, Any]
            the output from the git command
        """

        res_code, stdout, err = _git(['push'], self.repo_local_directory)

        if res_code != 0:
            raise Exception(err)

        return res_code, stdout, err



import os
import pprint
import shutil
from .version import __version__
from .git import Git
import warnings


# TODO: we should NOT be overwriting the samefile hook, but it'll work with GDrive for the meantime
# all files will be considered different
def samefile_network_hook(*args, **kwargs):
    return False


samefile_original_func = shutil._samefile


class Mizuna:

    def __init__(self,
                 repo_remote_url: str,
                 repo_local_directory: str,
                 networked_drive: bool = False,
                 verbose: bool = False):

        """
        Initialization method for Mizuna.

        Parameters
        ----------
        repo_remote_url: str
            Remote URL of the git repository
        repo_local_directory: str
            Local directory to maintain the git repository
        networked_drive: bool
            True if the local directory is a networked drive
        """

        self._cwd = os.getcwd()
        self.version = __version__

        print(f'Mizuna v{self.version}')
        print(f'Mizuna will setup a sync folder from your current working directory.')
        print(f'Current working directory: {self._cwd}')

        if networked_drive:
            warnings.warn(f'A bug in Python (see https://bugs.python.org/issue33935) prevents files in networked drives'
                          f'from copying properly.'
                          f'Mizuna will prevent a crucial routine in the copying method from running.'
                          f'This will allow the metadata of same files to be updated and overwritten.', RuntimeWarning)
            shutil._samefile = samefile_network_hook

        self._files_tracked = dict()
        self._files_tracked_count = 0
        self._mizuna_sync_folder = '.mizuna'

        if os.path.isdir(self._mizuna_sync_folder):
            print(f'Sync folder {self._mizuna_sync_folder} exists.')
        else:
            os.mkdir(self._mizuna_sync_folder)
            print(f'Sync folder {self._mizuna_sync_folder} does not exist -- creating.')

        print(f'Sync folder (relative): {self._mizuna_sync_folder}')
        print(f'Sync folder (absolute): {os.path.join(self._cwd, self._mizuna_sync_folder)}')

        self._repo_remote_url = repo_remote_url
        self._repo_local_directory = repo_local_directory

        print(f'Remote URL: {self._repo_remote_url}')
        print(f'Local directory: {self._repo_local_directory}')

        print('Connecting to git...')
        full_local_directory = os.path.join(self._mizuna_sync_folder, self._repo_local_directory)
        print(f'Sync local directory: {full_local_directory}')
        self._bridge = Git(repo_remote_url, full_local_directory, self._cwd)

        self._bridge.pull()

    def __str__(self):

        return_string = f'Repository Remote URL: {self._repo_remote_url}\n' \
                        f'Repository Local Directory: {self._repo_local_directory}'

        return return_string

    def track(self,
              file,
              remote_file: str = ''):

        # TODO: check if file exists before tracking? throw warning if files does not exist

        if isinstance(file, str):
            if isinstance(remote_file, str):
                self._add_single_track(file, remote_file)
            else:
                raise Exception('File remote location should be a string.')
        elif isinstance(file, dict):
            if remote_file != '':
                warnings.warn(f'Passing a dictionary of files, ignoring remote_file parameter.')
            self._add_multi_track(file)
        elif isinstance(file, list):
            if remote_file != '':
                warnings.warn(f'Passing a list of files, ignoring remote_file parameter.')
            for f in file:
                self._add_single_track(f, '')
        else:
            raise Exception('File location should be a string.')

        self._files_tracked_count = len(self._files_tracked)
        pprint.pprint(self._files_tracked)

    def _add_single_track(self,
                          file_path: str,
                          remote_path: str):

        if not os.path.exists(file_path):
            raise Exception(f'{file_path} does not exist')

        if file_path in self._files_tracked:
            warnings.warn(f'{file_path} is already being tracked, updating remote to: {remote_path}', RuntimeWarning)

        self._files_tracked[file_path] = remote_path

    def _add_multi_track(self,
                         files: dict):

        for k, v in files.items():
            if not isinstance(k, str):
                raise Exception(f'File location should be a string: {k}')
            if not isinstance(v, str):
                raise Exception(f'File remote location should be a string: {v}')

            try:
                self._add_single_track(k, v)
            except Exception as e:
                print(e)

    def untrack(self, file):
        self._files_tracked.pop(file)
        self._files_tracked_count = len(self._files_tracked)
        print(f'{file} untracked.')

    def untrack_all(self):
        self._files_tracked.clear()
        self._files_tracked_count = 0
        print(f'All files untracked.')

    def file_track_list(self):
        return self._files_tracked

    def sync(self):
        self._bridge.pull()

        if len(self._files_tracked) == 0:
            warnings.warn('Trying to call sync with no tracked files.', RuntimeWarning)
            return

        # copy files over sync folder
        # TODO: cleanup, remember the git add files should be relative to the local git repo
        for src, dst in self._files_tracked.items():
            if dst == '':
                dst = src
                dstn = os.path.join(self._bridge.repo_local_directory, src)
            else:
                dstn = os.path.join(self._bridge.repo_local_directory, dst)
            print(src, dstn)

            if not os.path.exists(os.path.dirname(dstn)):
                os.makedirs(os.path.dirname(dstn), exist_ok=True)
            shutil.copy2(src, dstn)

            res1 = self._bridge.add(dst)

        res2 = self._bridge.commit()
        res3 = self._bridge.push()

        return res1 or res2 or res3



import unittest
from unittest.mock import patch
import shutil
import os

from mizuna.mizuna import Mizuna


test_repo_url = 'https://git.overleaf.com/unittesturl'
test_repo_dir = 'UnitTestDir'
sync_dir_name = '.mizuna'

test_dir = 'tests'
file_single1 = 'figures/fig1.txt'
file_single2 = 'figures/fig2.txt'


class Utilities:

    @staticmethod
    def delete_sync_directory():
        if os.path.exists(sync_dir_name):
            shutil.rmtree(sync_dir_name)


class Initialization(unittest.TestCase):

    def setUp(self) -> None:
        Utilities.delete_sync_directory()

    def tearDown(self) -> None:
        Utilities.delete_sync_directory()

    @patch('mizuna.git.call_subprocess')
    def test_version(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        from mizuna.version import __version__
        m = Mizuna(test_repo_url, test_repo_dir)
        self.assertEqual(m.version, __version__)

    @patch('mizuna.git.call_subprocess')
    def test_initialization(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        m = Mizuna(test_repo_url, test_repo_dir)

    @patch('mizuna.git.call_subprocess')
    def test_clone_fail(self, mock_subprocess):
        mock_subprocess.return_value = (128, 'fail', 'fail')
        with self.assertRaises(Exception):
            m = Mizuna(test_repo_url, test_repo_dir)

    @patch('mizuna.git.call_subprocess')
    def test_create_sync_directory(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        m = Mizuna(test_repo_url, test_repo_dir)
        self.assertTrue(os.path.exists(sync_dir_name))

    @patch('mizuna.git.call_subprocess')
    def test_networked_drive_warning(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        with self.assertWarns(RuntimeWarning):
            m = Mizuna(test_repo_url, test_repo_dir, True)

    @patch('mizuna.git.call_subprocess')
    def test_print(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        m = Mizuna(test_repo_url, test_repo_dir)
        print(m)


class Adding(unittest.TestCase):

    @patch('mizuna.git.call_subprocess')
    def setUp(self, mock_subprocess) -> None:
        mock_subprocess.return_value = (0, 'mock', 'mock')
        self.m = Mizuna(test_repo_url, test_repo_dir)
        self.m.untrack_all()
        self.assertEqual(self.m._files_tracked_count, 0)

    def test_track_single(self):
        self.m.track(os.path.join(test_dir, file_single1), file_single1)
        self.assertEqual(self.m._files_tracked_count, 1)
        test_set = {os.path.join(test_dir, file_single1): file_single1}
        self.assertDictEqual(self.m.file_track_list(), test_set)

    def test_track_dict(self):
        self.m.track(os.path.join(test_dir, file_single1), file_single1)
        self.m.track(os.path.join(test_dir, file_single2), file_single2)
        self.assertEqual(self.m._files_tracked_count, 2)

        test_set = {os.path.join(test_dir, file_single1): file_single1,
                    os.path.join(test_dir, file_single2): file_single2}
        self.assertDictEqual(self.m.file_track_list(), test_set)

    def test_track_duplicate(self):
        with self.assertWarns(RuntimeWarning):
            self.m.track(os.path.join(test_dir, file_single1))
            self.m.track(os.path.join(test_dir, file_single1))

    def test_bad_track_local_single(self):
        with self.assertRaises(Exception):
            self.m.track(1234)

        self.assertEqual(self.m._files_tracked_count, 0)
        self.assertDictEqual(self.m.file_track_list(), dict())

    def test_bad_track_remote_single(self):
        with self.assertRaises(Exception):
            self.m.track(os.path.join(test_dir, file_single1), 1234)

        self.assertEqual(self.m._files_tracked_count, 0)
        self.assertDictEqual(self.m.file_track_list(), dict())

    def test_bad_track_dict(self):
        test_set = {file_single1: 1234,
                    file_single2: file_single2}
        with self.assertRaises(Exception):
            self.m.track(test_set)

        test_set = {file_single1: file_single1,
                    1234: file_single2}
        with self.assertRaises(Exception):
            self.m.track(test_set)

        self.assertEqual(self.m._files_tracked_count, 0)
        self.assertDictEqual(self.m.file_track_list(), dict())


class Removing(unittest.TestCase):

    test_set = {os.path.join(test_dir, file_single1): file_single1,
                os.path.join(test_dir, file_single1): file_single2}

    @patch('mizuna.git.call_subprocess')
    def setUp(self, mock_subprocess) -> None:
        mock_subprocess.return_value = (0, 'mock', 'mock')
        self.m = Mizuna(test_repo_url, test_repo_dir)
        self.m.untrack_all()
        self.assertEqual(self.m._files_tracked_count, 0)
        self.m.track(self.test_set)
        self.assertEqual(self.m._files_tracked_count, len(self.test_set))

    def test_remove_one(self):
        self.m.untrack(os.path.join(test_dir, file_single1))
        self.assertEqual(self.m._files_tracked_count, len(self.test_set) - 1)
        test_set_removed = self.test_set
        test_set_removed.pop(os.path.join(test_dir, file_single1))
        self.assertDictEqual(self.m.file_track_list(), test_set_removed)

    def test_remove_all(self):
        self.m.untrack_all()
        self.assertEqual(self.m._files_tracked_count, 0)
        self.assertDictEqual(self.m.file_track_list(), dict())


class Syncing(unittest.TestCase):

    @patch('mizuna.git.call_subprocess')
    def setUp(self, mock_subprocess) -> None:
        mock_subprocess.return_value = (0, 'mock', 'mock')
        self.m = Mizuna(test_repo_url, test_repo_dir)

    @patch('mizuna.git.call_subprocess')
    def test_sync(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        self.m.track(os.path.join(test_dir, file_single1), file_single1)
        result, stdout, err = self.m.sync()
        self.assertEqual(result, 0)
        self.assertTrue(os.path.exists(os.path.join(sync_dir_name, test_repo_dir, file_single1)))

    @patch('mizuna.git.call_subprocess')
    def test_sync_different_names(self, mock_subprocess):
        test_set = {os.path.join(test_dir, file_single1): file_single1,
                    os.path.join(test_dir, file_single2): ''}

        mock_subprocess.return_value = (0, 'mock', 'mock')
        self.m.track(test_set)
        result, stdout, err = self.m.sync()
        self.assertEqual(result, 0)
        self.assertTrue(os.path.exists(os.path.join(sync_dir_name, test_repo_dir, file_single1)))
        self.assertTrue(os.path.exists(os.path.join(sync_dir_name, test_repo_dir, 'tests', file_single2)))

    @patch('mizuna.git.call_subprocess')
    def test_sync_no_files(self, mock_subprocess):
        mock_subprocess.return_value = (0, 'mock', 'mock')
        with self.assertWarns(RuntimeWarning):
            self.m.sync()

    def tearDown(self) -> None:
        Utilities.delete_sync_directory()



from setuptools import setup, find_packages

v = {}
version = exec(open('mizuna/version.py').read(), v)
description = 'Mizuna is a package that automates uploading visualizations to Overleaf projects (or git repositories).'
long_description = open('README.md').read()

setup(
    name='mizuna',
    version=v['__version__'],
    author='Sebastian Rodriguez',
    author_email='s.rodriguez1850@outlook.com',
    url='https://github.com/srodriguez1850/Mizuna',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT License',
    packages=find_packages(),
    python_requires='>=3.6, <4',
    # install_requires=[],
    keywords=['python', 'workflow', 'data-science',
              'latex', 'overleaf',
              'jupyter',
              'matplotlib', 'seaborn', 'ggplot', 'plotly'],

    project_urls={
        'Bug Reports': 'https://github.com/srodriguez1850/Mizuna/issues',
        'Source': 'https://github.com/srodriguez1850/Mizuna/issues'
    },

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities'
    ]
)
